# lua-resty-openssl

FFI-based OpenSSL binding for LuaJIT, supporting OpenSSL 1.1 and 1.0.2 series

![Build Status](https://travis-ci.com/fffonion/lua-resty-openssl.svg?branch=master)


Table of Contents
=================

- [Description](#description)
- [Status](#status)
- [Synopsis](#synopsis)
- [Compatibility](#compatibility)
- [TODO](#todo)
- [Copyright and License](#copyright-and-license)
- [See Also](#see-also)


Description
===========

`lua-resty-openssl` is a FFI-based OpenSSL binding library, currently
supports OpenSSL `1.1.1`, `1.1.0` and `1.0.2` series.

The API is kept as same [luaossl](https://github.com/wahern/luaossl) while only a small sets
of OpenSSL APIs are currently implemented.


[Back to TOC](#table-of-contents)

Status
========

Production.

Synopsis
========

This library is greatly inspired by [luaossl](https://github.com/wahern/luaossl), while uses the
naming conversion closer to original OpenSSL API.
For example, a function called `X509_set_pubkey` in OpenSSL C API will expect to exist
as `resty.openssl.x509:set_pubkey`.
*CamelCase*s are replaced to *underscore_case*s, for exmaple `X509_set_serialNumber` becomes
`resty.openssl.x509:set_serial_number`.

Each Lua table returned by `new()` contains a cdata object `ctx`. User are not supposed to manully setting
`ffi.gc` or calling corresponding destructor of the `ctx` struct (like `*_free` functions).


## resty.openssl

This meta module provides a version sanity check and returns all exported modules to a local table.

```lua
local _M = {
  _VERSION = '0.1.0',
  version = require("resty.openssl.version"),
  pkey = require("resty.openssl.pkey"),
  digest = require("resty.openssl.digest"),
  bn = require("resty.openssl.bn"),
  x509 = require("resty.openssl.x509"),
  name = require("resty.openssl.x509.name"),
  altname = require("resty.openssl.x509.altname"),
  csr = require("resty.openssl.x509.csr"),
  extension = require("resty.openssl.x509.extension"),
}
```

### openssl.luaossl_compact

**syntax**: *openssl.luaossl_compact()*

Provides `luaossl` flavored API which uses *camelCase* naming; user can expect drop in replacement.

For example, `pkey:get_parameters` is mapped to `pkey:getParameters`.

## resty.openssl.version

A module to provide version info.

### version_num

The OpenSSL version number.

### OPENSSL_11

A boolean indicates whether the linked OpenSSL is 1.1 series.

### OPENSSL_10

A boolean indicates whether the linked OpenSSL is 1.0 series.

## resty.openssl.pkey

Module to provide EVP_PKEY infrastructure.

### pkey.new

**syntax**: *pk, err = pkey.new(config)*

**syntax**: *pk, err = pkey.new(string, format?)*

**syntax**: *pk, err = pkey.new()*

Creates a new pkey instance. The first argument can be:

1. A table which defaults to:

```lua
{
    type = 'RSA',
    bits = 2048,
    exp = 65537
}
```

to create EC private key:

```lua
{
    type = 'EC',
    curve = 'primve196v1',
}
```

2. A string of private or public key in PEM or DER format; optionally tells the library
to explictly decode the key using `format`, which can be a choice of `PER`, `DER` or `*`
for auto detect.
3. `nil` to create a 2048 bits RSA key.

### pkey.istype

**syntax**: *status = pkey.istype(table)*

Returns `true` if table is an instance of `pkey`. Returns `false` otherwise.

### pkey:get_parameters

**syntax**: *parameters, err = pk:get_parameters()*

Returns a table containing the `parameters` of pkey instance. Currently only `n`, `e` and `d`
parameter of RSA key is supported. Each value of the returned table is a
[resty.openssl.bn](#restyopensslbn) instance.

```lua
local pk, err = require("resty.openssl").pkey.new()
local parameters, err = pk:get_parameters()
local e = parameters.e
ngx.say(ngx.encode_base64(e:to_binary()))
-- outputs 'AQAB' (65537) by default
```

### pkey:sign

**syntax**: *signature, err = pk:sign(digest)*

Sign a [digest](#restyopenssldigest) using the private key defined in `pkey`
instance. The `digest` parameter must be a [resty.openssl.digest](#restyopenssldigest) 
instance. Returns the signed raw binary and error if any.

```lua
local pk, err = require("resty.openssl").pkey.new()
local digest, err = require("resty.openssl").digest.new("SHA256")
digest:update("dog")
local signature, err = pk:sign(digest)
ngx.say(ngx.encode_base64(signature))
```

### pkey:verify

**syntax**: *ok, err = pk:verify(signature, digest)*

Verify a signture (which can be generated by [pkey:sign](#pkey-sign)). The second
argument must be a [resty.openssl.digest](#restyopenssldigest) instance that uses
the same digest algorithm as used in `sign`.

### pkey:to_PEM

**syntax**: *pem, err = pk:to_PEM(private_or_public?)*

Outputs private key or public key of pkey instance in PEM format. `private_or_public`
must be a choice of `public`, `PublicKey`, `private`, `PrivateKey` or nil.
By default, it returns the public key.


## resty.openssl.bn

Module to expose BIGNUM structure.

### bn.new

**syntax**: *b, err = bn.new(bn_cdata?)*
**syntax**: *b, err = bn.new(number?)*

Creates a BIGNUM instance. The first argument can be `BIGNUM *` cdata object, or a Lua number,
or `nil` to creates a empty instance.

### bn.dup

### bn.from_binary

### bn.istype

**syntax**: *status = bn.istype(table)*

Returns `true` if table is an instance of `bn`. Returns `false` otherwise.

### bn:to_binary

**syntax**: *bin, err = bn:to_binary()*

Export the BIGNUM value in binary.

```lua
local b, err = require("resty.openssl.bn").new(23333)
local bin, err = b:to_binary()
ngx.say(ngx.encode_base64(bin))
-- outputs "WyU="
```

## resty.openssl.digest

Module to interact with message digest.

### digest.new

**syntax**: *d, err = digest.new(digest_name)*

Creates a digest instance.`digest_name` is a string of digest algorithm name. To view
a list of digest algorithms implemented, use `openssl list -digest-algorithms`.

### digest.istype

**syntax**: *status = digest.istype(table)*

Returns `true` if table is an instance of `digest`. Returns `false` otherwise.

### digest:update

**syntax**: *digest:update(partial, ...)*

Updates the digest with one or more string.

### digest:final

**syntax**: *digest:update(partial?, ...)*

## resty.openssl.rand

### rand.bytes

## resty.openssl.x509

Module to interact with certificate.

### x509.new

### x509.istype

**syntax**: *status = x509.istype(table)*

Returns `true` if table is an instance of `x509`. Returns `false` otherwise.

### x509:add_extension

**syntax**: *status = x509:add_extension(extension)*

### x509:get_issuer_name

### x509:set_issuer_name

### x509:get_lifetime

### x509:set_lifetime

### x509:get_not_before

### x509:set_not_before

### x509:get_not_after

### x509:set_not_after

### x509:get_pubkey

### x509:set_pubkey

### x509:get_serial_number

### x509:set_serial_number

### x509:get_subject_name

### x509:set_subject_name

### x509:get_version

### x509:set_version

### x509:set_basic_constraints

### x509:set_basic_constraints_critical

### x509:sign

## resty.openssl.x509.altname

### altname.new

### altname.istype

**syntax**: *altname = digest.istype(table)*

Returns `true` if table is an instance of `altname`. Returns `false` otherwise.

### altname:add

## resty.openssl.x509.csr

Module to interact with certificate signing request.

### csr.new

### csr.istype

**syntax**: *status = csr.istype(table)*

Returns `true` if table is an instance of `csr`. Returns `false` otherwise.

### csr:set_subject_name

### csr:set_subject_alt

### csr:set_pubkey

### csr:tostring

### csr:to_PEM

## resty.openssl.x509.extension

### extension.new

### extension.istype

**syntax**: *status = extension.istype(table)*

## resty.openssl.x509.name

### name.new

### name.dup

### name.istype

**syntax**: *status = name.istype(table)*

Returns `true` if table is an instance of `name`. Returns `false` otherwise.

### name:add

**syntax**: *name, err = name:add(nid, txt)*

Adds an ASN.1 object to `name`. First arguments in the *text representation* of
[NID](https://boringssl.googlesource.com/boringssl/+/HEAD/include/openssl/nid.h).
Second argument is the plain text value for the ASN.1 object.

Returns the name instance itself on success, or `nil` and an error on failure.

Compatibility
====

Although only a small combinations of CPU arch and OpenSSL version are tested, the library
should function well as long as the linked OpenSSL library is API compatible. This means
the same name of functions are exported with same argument types.

For OpenSSL 1.0.2 series however, ABI compatibility must be ensured as some struct members
are accessed directly, meaning they are accessed by memory offset in assembly.

If you plan to use this library on an untested version of OpenSSL (like custom builds or pre releases),
[this](https://abi-laboratory.pro/index.php?view=timeline&l=openssl) can be a good source to consult.

TODO
====

- review get0 function calls to ensure there's no double free
- find a way to test memory leak
- seperate cdef with lua implementation to allow cleaner dependency

[Back to TOC](#table-of-contents)


Copyright and License
=====================

This module is licensed under the BSD license.

Copyright (C) 2019, by fffonion <fffonion@gmail.com>.

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

[Back to TOC](#table-of-contents)

See Also
========
* [luaossl](https://github.com/wahern/luaossl)
* [API/ABI changes review for OpenSSL](https://abi-laboratory.pro/index.php?view=timeline&l=openssl)

[Back to TOC](#table-of-contents)
